
package require snit

package provide compass_utils 1.1

################################################################################
#   Small utilities
################################################################################

namespace eval ::math::constants {}
proc ::math::constants::give_constant { name } {
    ::math::constants::constants $name
    return [set $name]
}

namespace eval cu {}

proc ::_ { args } {
    set ret [uplevel 1 ::msgcat::mc $args]
    regexp {(.*)#C#(.*)} $ret {} ret
    return $ret
}


proc range {from to {step 1}} {
    set res ""
    if { $step > 0 } {
	for { set i $from } { $i < $to } { set i [expr {$i+$step}] } {
	    lappend res $i
	}
    } else {
	for { set i $from } { $i > $to } { set i [expr {$i+$step}] } {
	    lappend res $i
	}
    }
    return $res
}

proc rangeF {from to {step 1}} {
    set res ""
    if { $step > 0 } {
	for { set i $from } { $i <= $to } { set i [expr {$i+$step}] } {
	    lappend res $i
	}
    } else {
	for { set i $from } { $i >= $to } { set i [expr {$i+$step}] } {
	    lappend res $i
	}
    }
    return $res
}

if { [info command lreverse] eq "" } {
    proc lreverse { L } {
	set res {}
	set i [llength $L]
	#while {[incr i -1]>=0} {lappend res [lindex $L $i]}
	while {$i} {lappend res [lindex $L [incr i -1]]} ;# rmax
	return $res
    }
}

proc list_unique { list } {
    return [lsort -dictionary -unique $list]
}

proc list_union { list1 list2 } {
    return [lsort -dictionary -unique [concat $list1 $list2]]
}

proc list_intersection { list1 list2 } {
    set nlist ""
    foreach i $list2 {
	if { [lsearch -exact $list1 $i] != -1 } {
	    lappend nlist $i
	}
    }
    return $nlist
}

proc list_difference { list1 list2 } {
    set nlist ""
    foreach i $list1 {
	if { [lsearch -exact $list2 $i] == -1 } {
	    lappend nlist $i
	}
    }
    return $nlist
}

proc dict_inverse { dict } {
    set ret ""
    dict for "n v" $dict {
	dict set ret $v $n
    }
    return $ret
}

proc linsert0 { args } {
    set optional {
	{ -max_len len "" }
    }
    set compulsory "list element"
    parse_args $optional $compulsory $args

    set ipos [lsearch -exact $list $element]
    if { $ipos != -1 } {
	set list [lreplace $list $ipos $ipos]
    }
    set list [linsert $list 0 $element]
    if { $max_len ne "" } {
	set list [lrange $list 0 $max_len]
    }
    return $list
}

# regmap re string var script
# substitute all the occurrences of 're' in 'string' with
# the result of the evaluation of 'script'. Script is called
# for every match with 'var' variable set to the matching string.
#
# Example: regmap {[0-9]+} "1 2 hello 3 4 world" x {expr {$x+1}}
# Result:  "2 3 hello 4 5 world"
proc regmap {re string var script} {
    set submatches [lindex [regexp -about $re] 0]
    lappend varlist idx
    while {[incr submatches -1] >= 0} {
	lappend varlist _
    }
    set res $string
    set indices [regexp -all -inline -indices $re $string]
    set delta 0
    foreach $varlist $indices {
	foreach {start end} $idx break
	set substr [string range $string $start $end]
	uplevel [list set $var $substr]
	set subresult [uplevel $script]
	incr start $delta
	incr end $delta
	set res [string replace $res $start $end $subresult]
	incr delta [expr {[string length $subresult]-[string length $substr]}]
    }
    return $res
}

proc regsubfile { args } {

    set opts ""
    set noopts ""
    set isopt 0
    foreach i $args {
	if { $isopt || ![string match -* $i] } {
	    lappend noopts $i
	    set isopt 1
	} else {
	    lappend opts $i
	    if { $i == "--" } { set isopt 1 }
	}
    }
    if { [llength $noopts] == 4 } {
	foreach "exp file subSpec varName" $noopts break
    } elseif { [llength $noopts] == 3 } {
	foreach "exp file subSpec" $noopts break
    } else { error "error. usage: regsubfile ?switches? exp file subSpec ?varName?" }
    
    set fin [open $file r]
    set data [read $fin]
    close $fin
    set retval [eval regsub $opts [list $exp $data $subSpec data_out]]

    if { [info exists varName] } {
	upvar $varName x
	set x $data_out
    }

    #file copy -force $file $file~
    if { $data ne $data_out } {
	set fout [open $file w]
	puts -nonewline $fout $data_out
	close $fout
    } else { set retval 0 }
    return $retval
}


proc dict_getd { dictionaryValue key default } {
    if { [dict exists $dictionaryValue $key] } {
	return [dict get $dictionaryValue $key]
    }
    return $default
}

# prints a tree defined as struct::tree
proc print_tree { tree { node root } } {
    set _ ""
    $tree walk $node i {
	append _ "$i [$tree parent $i] [$tree getall $i]\n"
    }
    return $_
}

proc print_tree_xml { tree { node root } } {
    package require tdom
    set _ ""
    $tree walk $node -order both "action i" {
	if { $action eq "enter" } {
	    append _ "<$i "
	    foreach "n v" [$tree getall $i] {
		if { [string is integer $n] } {
		    set n integer_$n
		}
		if { [string is double -strict $v] } {
		    set v [format %.3g $v]
		}
		append _ "$n='[xml_map $v]' "
	    }
	    append _ ">"
	} else {
	    append _ "</$i>"
	}
    }
    dom parse $_ doc
    
    package require RamDebugger
    interp eval ramdebugger [list RamDebugger::OpenFileSaveHandler \
	    "*formula (xml)*" [$doc asXML] ""]
    
}

namespace eval cu::string {}

proc cu::string::maintain_case { word case_word } {
    if { $word ne [string tolower $word] } { return $word }
    if { $case_word eq [string toupper $case_word] } { return [string toupper $word] }
    set cw0 [string index $case_word 0]
    if { $cw0 eq [string toupper $cw0] } { return [string totitle $word] }
    return $word
}

namespace eval cu::file {
    variable tempdir    {}
    variable tempdirSet 0
}

proc cu::file::appdatadir { program_name } {
    if { [info exists ::env(APPDATA)] } {
	set AppDataDir [file join $::env(APPDATA) $program_name]
    } elseif { $::tcl_platform(platform) == "windows" } {
	package require registry
	set key {HKEY_CURRENT_USER\Software\Microsoft\Windows\CurrentVersion}
	append key {\Explorer\Shell Folders}
	set err [catch { registry get $key AppData } AppData]
	if { !$err } {
	    set AppDataDir [file join [registry get $key AppData] $program_name]
	} else {
	    set AppDataDir [file join $::env(HOME) .$program_name]
	}
    } else { set AppDataDir [file join $::env(HOME) .$program_name] }
    
    file mkdir $AppDataDir
    
    return $AppDataDir
}

proc cu::file::tempdir {args} {
    if {[llength $args] > 1} {
	return -code error {wrong#args: should be "cu::file::tempdir ?path?"}
    } elseif {[llength $args] == 1} {
	variable tempdir    [lindex $args 0]
	variable tempdirSet 1
	return
    }
    return [file normalize [_tempdir]]
}

proc cu::file::_tempdir {} {
    global tcl_platform env
    variable tempdir
    variable tempdirSet

    set attempdirs [list]

    if {$tempdirSet} {
	lappend attempdirs $tempdir
    } else {
	foreach tmp {TMPDIR TEMP TMP} {
	    if { [info exists env($tmp)] } {
		lappend attempdirs $env($tmp)
	    }
	}

	switch $tcl_platform(platform) {
	    windows {
		lappend attempdirs "C:\\TEMP" "C:\\TMP" "\\TEMP" "\\TMP"
	    }
	    macintosh {
		set tmpdir $env(TRASH_FOLDER)  ;# a better place?
	    }
	    default {
		lappend attempdirs [file join / tmp] \
		        [file join / var tmp] [file join / usr tmp]
	    }
	}

	lappend attempdirs [pwd]
    }

    foreach tmp $attempdirs {
	if { [file isdirectory $tmp] && [file writable $tmp] } {
	    return $tmp
	}
    }

    # Fail if nothing worked.
    return -code error "Unable to determine a proper directory for temporary files"
}

proc cu::file::tempfile {{prefix {}}} {
    return [file normalize [_tempfile $prefix]]
}

proc cu::file::_tempfile {prefix} {
    set tmpdir [tempdir]

    set chars "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    set nrand_chars 10
    set maxtries 10
    set access [list RDWR CREAT EXCL TRUNC]
    set permission 0600
    set channel ""
    set checked_dir_writable 0
    set mypid [pid]
    for {set i 0} {$i < $maxtries} {incr i} {
	 set newname $prefix
	 for {set j 0} {$j < $nrand_chars} {incr j} {
	     append newname [string index $chars \
		    [expr {int(rand()*62)}]]
	 }
	set newname [file join $tmpdir $newname]
	 if {[file exists $newname]} {
	     after 1
	 } else {
	     if {[catch {open $newname $access $permission} channel]} {
		 if {!$checked_dir_writable} {
		     set dirname [file dirname $newname]
		     if {![file writable $dirname]} {
		         return -code error "Directory $dirname is not writable"
		     }
		     set checked_dir_writable 1
		 }
	     } else {
		 # Success
		close $channel
		return $newname
	     }
	 }
    }
    if {[string compare $channel ""]} {
	 return -code error "Failed to open a temporary file: $channel"
    } else {
	 return -code error "Failed to find an unused temporary file name"
    }
}

proc cu::file::make_path_relative { basedir path } {
    set basedirList [file split [file normalize $basedir]]
    set pathList [file split [file normalize $path]]
    
    set len1 [llength $basedirList]
    set len2 [llength $pathList]
    for { set i 0 } { $i < $len1 && $i < $len2 } { incr i } {
	if { [lindex $basedirList $i] ne [lindex $pathList $i] } { break }
    }
    if { $i < 1 } { return $path }
    set newpath [string repeat ../ [expr {$len1-$i}]]
    set newpath [eval [list file join $newpath] [lrange $pathList $i end]]
    return $newpath
}

# cu::file::stripPath (copied from ::fileutil::stripPath )
#
#        If the specified path references/is a path in prefix (or prefix itself) it
#        is made relative to prefix. Otherwise it is left unchanged.
#        In the case of it being prefix itself the result is the string '.'.
#
# Arguments:
#        prefix                prefix to strip from the path.
#        path                path to modify
#
# Results:
#        path                The (possibly) modified path.

proc cu::file::stripPath {prefix path} {
    # [file split] is used to generate a canonical form for both
    # paths, for easy comparison, and also one which is easy to modify
    # using list commands.

    if {[string equal $prefix $path]} {
	return "."
    }

    set prefix [file split $prefix]
    set npath  [file split $path]

    if {[string match ${prefix}* $npath]} {
	set path [eval [linsert [lrange $npath [llength $prefix] end] 0 file join ]]
    }
    return $path
}

# cu::file::jail (copied and fixed from fileutil::jail)
#
#        Ensures that the input path 'filename' stays within the the
#        directory 'jail'. In this way it preventsuser-supplied paths
#        from escaping the jail.
#
# Arguments:
#        jail                The path to the directory the other must
#                        not escape from.
#        filename        The path to prevent from escaping.
#
# Results:
#        path                The (possibly) modified path surely within
#                        the confines of the jail.

proc cu::file::jail {jail filename} {
    if {![string equal [file pathtype $filename]  "relative"]} {
	# Although the path to check is absolute (or volumerelative on
	# windows) we cannot perform a simple prefix check to see if
	# the path is inside the jail or not. We have to normalize
	# both path and jail and then we can check. If the path is
	# outside we make the original path relative and prefix it
	# with the original jail. We do make the jail pseudo-absolute
	# by prefixing it with the current working directory for that.

	# Normalized jail. Fully resolved sym links, if any. Our main
	# complication is that normalize does not resolve symlinks in the
	# last component of the path given to it, so we add a bogus
	# component, resolve, and then strip it off again. That is why the
	# code is so large and long.

	set njail [eval [list file join] [lrange [file split \
		[file normalize [file join $jail __dummy__]]] 0 end-1]]

	# Normalize filename. Fully resolved sym links, if
	# any. S.a. for an explanation of the complication.

	set nfile [eval [list file join] [lrange [file split \
		[file normalize [file join $filename __dummy__]]] 0 end-1]]

	if {[string match ${njail}* $nfile]} {
	    return $filename
	}

	# Outside the jail, put it inside. ... We normalize the input
	# path lexically for this, to prevent escapes still lurking in
	# the original path. (We cannot use the normalized path,
	# symlinks may have bent it out of shape in unrecognizable ways.

	return [eval [linsert [lrange [file split \
		[_lexnormalize $filename]] 1 end] 0 file join [pwd] $jail]]
    } else {
	# The path is relative, consider it as outside
	# implicitly. Normalize it lexically! to prevent escapes, then
	# put the jail in front, use PWD to ensure absoluteness.

	return [eval [linsert [file split [_lexnormalize $filename]] 0 \
		file join [pwd] $jail]]
    }
}

proc cu::file::_lexnormalize {sp} {
    set sp [file split $sp]

    # Resolution of embedded relative modifiers (., and ..).

    set np {}
    set noskip 1
    while {[llength $sp]} {
	set ele    [lindex $sp 0]
	set sp     [lrange $sp 1 end]
	set islast [expr {[llength $sp] == 0}]

	if {[string equal $ele ".."]} {
	    if {[llength $np] > 1} {
		# .. : Remove the previous element added to the
		# new path, if there actually is enough to remove.
		set np [lrange $np 0 end-1]
	    }
	} elseif {[string equal $ele "."]} {
	    # Ignore .'s, they stay at the current location
	    continue
	} else {
	    # A regular element.
	    lappend np $ele
	}
    }
    if {[llength $np] > 0} {
	return [eval file join $np]
    }
    return {}
}

proc cu::file::execute { what file args } {
    switch -- $what {
	gid {
	    if { [file executable "~/GiD Project/gidvs/Release/gid.exe"] } {
		set pwd [pwd]
		cd "~/GiD Project"
		eval exec gidvs/Release/gid.exe [list $file] $args &
		#exec /home/ramsan/bin/gid &
		cd $pwd
	    } else {
		tk_messageBox -message [_ "Could not find a GiD executable"]
	    }
	}
	emacs {
	    exec runemacs -g 100x72 &
	}
	wish {
	    set pwd [pwd]
	    cd [file dirname $file]
	    eval exec wish [list [file normalize $file]] $args &
	    cd $pwd
	}
	tkdiff {
	    set pwd [pwd]
	    cd [file dirname $file]
	    exec wish ~/myTclTk/tkcvs/bin/tkdiff.tcl -r [file tail $file] &
	    cd $pwd
	}
	start {
	    if { [regexp {[&]} $file] } {
		set bat [file join [file dirname $file] a.bat]
		set fout [open $bat w]
		puts $fout "start \"\" \"$file\""
		close $fout
		exec $bat 
		file delete $bat
	    } else {
		eval exec [auto_execok start] \"\" [list $file] &
	    }
	}
	url {
	    if { ![regexp {(?i)^\w+://} $file] && ![regexp {(?i)^mailto:} $file] } {
		set txt [_ "url does not begin with a known handler like: %s. Proceed?" \
		        "http:// ftp:// mailto:"]
		set retval [tk_messageBox -default ok -icon question -message $txt \
		        -type okcancel]
		if { $retval == "cancel" } { return }
	    }
	    exec rundll32 url.dll,FileProtocolHandler $file &
	}
	default {
	    eval exec [list $file] $args &
	}
    }  
}

proc date_sql2tcl { date } {
    if { [regexp {^[-: 0.]*$} $date] } { return "" }
    return [clock scan $date]
}
proc date_tcl2sql { date } {
    if { $date == "" } { return "" }
    return [clock format $date -format "%Y-%m-%d %H:%M:%S"]
}

proc cu::bgerror { err } {
    set info $::errorInfo
    destroy ._bgerror
    set w [dialogwin_snit ._bgerror -title [_ "Warning"] -entrytext \
	    $err -cancelname -]
    set cmd "[list set ::errorInfo $info];"
    append cmd [list ::tk::dialog::error::bgerror $err]
    bind $w <Control-x><Control-x> $cmd
    set action [$w createwindow]
    destroy $w
}

proc cu::init_release_bgerror {} {
    auto_load ::bgerror
    interp alias "" ::bgerror "" cu::bgerror
}


################################################################################
#    Package to make logging for the debug phase
################################################################################

namespace eval ::mylog {
    variable sqlite3explorer {C:\TclTk\sqlite-3.2.1-sources\sqlite3explorer\sqlite3explorer.exe}
    variable logfile ""

    namespace export mylog
}

# if levels is void, disconnected
# levels can contain: debug debugoff log error
proc ::mylog::init { args } {
    variable logfile
    
    set optional {
	{ -view_binding binding "" }
	{ -file file "" }
	{ -restart_file boolean 0 }
    }
    set compulsory { levels }

    parse_args $optional $compulsory $args

    set logfile $file

    if { [llength $levels] } {
	package require sqlite3
	if { $logfile ne "" } {
	    if { $restart_file } {
		file delete $file
	    }
	    sqlite3 mylog::db $logfile
	} else {
	    sqlite3 mylog::db :memory:
	}
	mylog::db eval { PRAGMA synchronous = OFF }
	if { ![mylog::table_exists mylog::db log] } {
	    mylog::db eval { create table log(idx integer primary key,date datetime,
		level text,txt text,proc text,proc_args text) }
	}
    }
    foreach i [list debug debugoff log error] {
	if { [lsearch -exact $levels $i] != -1 } {
	    interp alias "" ::mylog::$i "" ::mylog::mylog $i
	} else {
	    proc ::mylog::$i { args } {}
	}
    }
    if { [llength $levels] && $view_binding ne "" } {
	bind all $view_binding [list mylog::view_log]
    }
}

proc ::mylog::table_exists { db name } {
    
    set ret [$db eval { select name from sqlite_master where type='table' 
	    and name=$name }]
    if { ![llength $ret] } { return 0 } else { return 1 }
}

proc ::mylog::mylog { level txt } {
    set now [clock clicks -milliseconds]
    set mili [expr {$now%1000}]
    set secs [expr {$now/1000}]
    set format "%Y-%m-%d %H:%M:%S.$mili"
    set date [clock format $secs -format $format]
    set err [catch { info level -1 } ret]
    if { !$err && [lsearch -exact "debug debugoff" $level] != -1 } {
	set proc [lindex $ret 0]
	set proc_args [lrange $ret 1 end]
    } else { foreach "proc proc_args" [list "" ""] break }
    mylog::db eval { insert into log (date,level,txt,proc,proc_args) 
	values($date,$level,$txt,$proc,$proc_args) }
}

proc ::mylog::clear_log { args } {

    set optional {
	{ -confirm boolean 0 }
	{ -parent parent_window . }
    }
    parse_args $optional "" $args

    if { $confirm } {
	set num [mylog::db onecolumn { select count(*) from log }]
	set text [_ "There are %s message. Delete all of them?" $num]
	set retval [snit_messageBox -default ok -icon question -message $text \
		-type okcancel -parent $parent]
	if { $retval == "cancel" } { return }
    }
    mylog::db eval { delete from log }
}

proc ::mylog::view_log { args } {
    variable sqlite3explorer
    variable logfile

    set optional {
	{ -extlogfile file "" }
	{ -parent parent_window "" }
    }
    parse_args $optional "" $args

    if { $extlogfile ne "" } {
	set file $extlogfile
    } else {
	set file $logfile
    }
    set err [catch {
	    package require tile
	    package require fulltktree
	    package require dialogwin
	}]
    if { !$err } {
	eval _view_log_fulltktree $args
	return
    }
    if { $file eq "" } {
	set file [file join $::env(TEMP) logfile.db3]
	
	file delete $file
	sqlite3 mylog::other $file
	mylog::db eval {SELECT sql FROM sqlite_master WHERE sql NOT NULL} {
	    mylog::other eval $sql
	}
	mylog::other close
	
	mylog::db eval "ATTACH '$file' AS app"
	mylog::db eval {SELECT name FROM sqlite_master WHERE type='table'} {
	    mylog::db eval "INSERT INTO app.$name SELECT * FROM $name"
	}
	mylog::db eval {DETACH app}
    }
    exec $sqlite3explorer $file &
}

proc ::mylog::_view_log_fulltktree { args } {

    set optional {
	{ -extlogfile file "" }
	{ -parent parent_window "" }
    }
    parse_args $optional "" $args

    package require sqlite3
    package require fulltktree
    package require dialogwin

    set w [dialogwin_snit $parent.%AUTO% -title Log -okname - \
	    -cancelname [_ Close] -morebuttons [list [_ Clear]]]
    set f [$w giveframe]

    set cols {
	{ 10 idx left item 0 }
	{ 10 date right text 0 }
	{ 8 level left text 0 }
	{ 30 txt left text 0 }
	{ 8 proc left text 0 }
	{ 8 proc_args left text 0 }
    }
    fulltktree $f.t -columns $cols -width 600 -height 600 \
	-compass_background 1 -showlines 0 \
	-indent 0 -expand 0 -sensitive_cols all \
	-selecthandler2 [list mylog::_view_log_row]

    grid $f.t -sticky nsew -padx 2 -pady 2
    grid columnconfigure $f 0 -weight 1
    grid rowconfigure $f 0 -weight 1

    if { $extlogfile ne "" } {
	sqlite3 mylog::db10 $extlogfile
	set db mylog::db10
	$w disablebutton 2
    } else {
	set db mylog::db
    }
    $db eval "select * from log order by idx desc" {
	$f.t insert end [list $idx $date $level $txt $proc $proc_args]
    }
    if { $extlogfile ne "" } {
	$db close
    }
    set action [$w createwindow]
    if { $action == 2 } {
	$db eval "delete from log"
	destroy $w
	eval ::mylog::_view_log_fulltktree $args
    } else {
	destroy $w
    }
}

proc ::mylog::_view_log_row { tree ids } {

    set id [lindex $ids 0]
    set w [dialogwin_snit $tree.%AUTO% -title [_ "Log"] -okname -]
    set f [$w giveframe]
    
    set text [text $f.t -xscrollcommand [list $f.sh set] -yscrollcommand \
	    [list $f.sv set]]
    scrollbar $f.sv -orient vertical -command [list $f.t yview]
    scrollbar $f.sh -orient horizontal -command [list $f.t xview]

    set idx 0
    foreach col [[winfo parent $tree] cget -columns] {
	$text insert end "[lindex $col 1]: " bold
	$text insert end [$tree item text $id $idx]\n\n
	incr idx
    }
    $text tag configure bold -font [concat [font actual [$f.t cget -font]] \
	    -weight bold]
    $f.t configure -state disabled
    bind $f.t <1> [list focus $f.t]

    grid $f.t $f.sv -sticky wns
    grid $f.sh -sticky new

    grid configure $f.t -sticky nsew
    grid columnconfigure $f 0 -weight 1
    grid rowconfigure $f 0 -weight 1
    $w createwindow
    destroy $w
}

################################################################################
#    instrument times in procs for debugging
################################################################################

namespace eval ::stime {
    variable tree_times
}

proc ::stime::init {} {
    variable tree_times

    package require struct::tree
    set tree_times [::struct::tree]
}

proc ::stime::register { args } {

    set cmd {
	foreach "procname text" %ARGS% {
	    set ns [namespace which -command $procname]
	    set newprocname [namespace qualifiers $ns]::_[namespace tail $procname]
	    if { [info command $procname] eq "" } {
		error "error in stime::register. proc '$procname' does not exist"
	    }
	    rename $procname $newprocname
	    proc $procname { args } "stime::add_time \" $newprocname \$args \" {$text} 2"
	}
    }
    uplevel 1 [string map [list %ARGS% [list $args]] $cmd]
}

proc ::stime::add_time { script namestack { level 1 } } {
    variable tree_times

    set ret [time { set retval [uplevel $level $script] }]
    set t [expr {[lindex $ret 0]*1e-6}]

    set cnode root
    foreach n $namestack {
	set found 0
	foreach node [$tree_times children $cnode] {
	    if { [$tree_times get $node name] eq $n } {
		set found 1
		break
	    }
	}
	if { $found } {
	    set cnode $node
	} else {
	    set cnode [$tree_times insert $cnode end]
	    $tree_times set $cnode name $n
	}
    }
    if { ![$tree_times keyexists $cnode t] } {
	$tree_times set $cnode t $t
    } else {
	$tree_times set $cnode t [expr {$t+[$tree_times set $cnode t]}]
    }
    $tree_times set $cnode userdef 1
    return $retval
}

proc ::stime::print { { channel "" } } {
    variable tree_times

    set _ ""
    $tree_times walk root -type bfs -order post i {
	set parent [$tree_times parent $i]
	if { $i eq "root" || $parent eq "root" } { continue }
	if { ![$tree_times keyexists $parent userdef] } {
	    if { ![$tree_times keyexists $parent t] } {
		set tp 0.0
	    } else {
		set tp [$tree_times set $parent t]
	    }
	    $tree_times set $parent t [expr {$tp+[$tree_times set $i t]}]
	}
    }
    $tree_times walk root i {
	if { $i eq "root" } { continue }
	set parent [$tree_times parent $i]
	if { $parent eq "root" } {
	    set percent 100
	} else {
	    set percent [expr {[$tree_times set $i t]*100.0/[$tree_times set $parent t]}]
	}
	append _ [_print_line [$tree_times depth $i] [$tree_times set $i name] \
		      [$tree_times set $i t] $percent]
    }
    if { $channel ne "" } { puts $channel $_ }
    return $_
}

proc ::stime::_print_line { level name time percent } {

    set _ [string repeat ..... [expr {$level-1}]]
    append _ [format "%12s %8.3fmin = %8.3fs (%3.2f%%)" $name [expr {$time/60.0}] \
		  $time $percent] "\n"
    return $_
}

################################################################################
#    parse args
#
# example:
#  proc myproc { args } {
#     set optional {
#         { -view_binding binding "" }
#         { -file file "" }
#         { -restart_file boolean 0 }
#         { -flag1 "" 0 }
#     }
#     set compulsory "levels"
#     parse_args $optional $compulsory $args
#
#     if { $view_binding ne "" } { puts hohoho }
#     if { $flag1 } { puts "activated flag" }
#  }
# 
################################################################################

proc ::parse_args { args } {

    set optional {
	{ -raise_compulsory_error boolean 1 }
	{ -compulsory_min min_number "" }
    }
    set compulsory "optional compulsory arguments"

    set cmdname [lindex [info level [expr {[info level]-1}]] 0]

    if { [string match -* [lindex $args 0]] } {
	parse_args $optional $compulsory $args
    } else {
	set raise_compulsory_error 1
	set compulsory_min ""
	if { [llength $args] != [llength $compulsory] } {
	    uplevel 1 [list error [_parse_args_string $cmdname $optional \
		        $compulsory $args]]
	    return ""
	}
	foreach $compulsory $args break
    }

    foreach i $optional {
	foreach "name namevalue default" $i break
	set opts_value($name) $namevalue
	set opts($name) $default
    }
    while { [string match -* [lindex $arguments 0]] } {
	if { [lindex $arguments 0] eq "--" } { break }
	foreach "name value" [lrange $arguments 0 1] break
	if { [regexp {(.*)=(.*)} $name {} name value] } {
	    set has_att_value 1
	} else {
	    set has_att_value 0
	}
	if { [info exists opts($name)] } {
	    if { $has_att_value } {
		set opts($name) $value
		set arguments [lrange $arguments 1 end]
	    } elseif { $opts_value($name) eq "" } {
		set opts($name) 1
		set arguments [lrange $arguments 1 end]
	    } else {
		set opts($name) $value
		set arguments [lrange $arguments 2 end]
	    }
	} else {
	    uplevel 1 [list error [_parse_args_string $cmdname $optional \
		        $compulsory $args]]
	    return ""
	}
    }
    if { $raise_compulsory_error } {
	if { $compulsory_min ne "" } {
	    if { [llength $arguments] < $compulsory_min || \
		[llength $arguments] > [llength $compulsory] } {
		uplevel 1 [list error [_parse_args_string $cmdname $optional $compulsory]]
		return ""
	    }
	} elseif { [llength $arguments] != [llength $compulsory] } {
	    uplevel 1 [list error [_parse_args_string $cmdname $optional \
		        $compulsory $args]]
	    return ""
	}

    }
    foreach name [array names opts] {
	uplevel 1 [list set [string trimleft $name -] $opts($name)]
    }
    set inum 0
    foreach i $compulsory {
	uplevel 1 [list set $i [lindex $arguments $inum]]
	incr inum
    }
    return [lrange $arguments $inum end]
}

proc ::_parse_args_string { cmd optional compulsory arguments } {

    set str "error. usage: $cmd "
    foreach i $optional {
	foreach "name namevalue default" $i break
	append str "?$name $namevalue? "
    }
    append str $compulsory
    append str "\n\targs: $arguments"
    return $str
}

################################################################################
#    gestures
#
################################################################################

namespace eval gestures {
    variable path
    variable callback
}

proc gestures::init { w _callback } {
    variable callback 

    set callback($w) $_callback

    _prepare_for_gestures $w gestures_$w
    bind gestures_$w <ButtonPress-3> [list gestures::press %X %Y]
    bind gestures_$w <B3-Motion> [list gestures::motion %X %Y]
    bind gestures_$w <ButtonRelease-3> [list gestures::release $w %X %Y]
}

proc gestures::add { w wtree } {
    _prepare_for_gestures $wtree gestures_$w
}

proc gestures::addT { wtree } {
    set w [winfo toplevel $wtree]
    _prepare_for_gestures $wtree gestures_$w
}

proc gestures::_prepare_for_gestures { w bindtag } {
    bindtags $w [linsert [bindtags $w] 0 $bindtag]
    foreach i [winfo children $w] { _prepare_for_gestures $i $bindtag }
}

proc gestures::press { x y } {
    variable path

    set path [list $x $y]
}

proc gestures::motion { x y } {
    variable path

    lappend path $x $y
}

proc gestures::release { w x y } {
    variable path
    variable callback

    lappend path $x $y

    set t 10
    foreach "x y" $path {
	if { ![info exists box(minx)] || $x < $box(minx) } { set box(minx) $x }
	if { ![info exists box(miny)] || $y < $box(miny) } { set box(miny) $y }
	if { ![info exists box(maxx)] || $x > $box(maxx) } { set box(maxx) $x }
	if { ![info exists box(maxy)] || $y > $box(maxy) } { set box(maxy) $y }
    }

    if { $box(maxx)-$box(minx) < $t && $box(maxy)-$box(miny) < $t } { return }

    foreach "x0 y0" [lrange $path 0 1] break
    foreach "xend yend" [lrange $path end-1 end] break

    if { $x0-$xend >= $t } {
	set gesture GestureLeft
    } elseif { $xend-$x0 >= $t } {
	set gesture GestureRight
    } elseif { $x0-$box(miny) >= $t && $xend-$box(miny) >= $t } {
	set gesture GestureUp,GestureDown
    } else { return }
   
    uplevel #0 $callback($w) $gesture
    return -code break
}

################################################################################
#    XML & xpath utilities
#
################################################################################

proc xml_map { args } {
    set map [list > "&gt;" < "&lt;" & "&amp;" \" "&quot;" ' "&apos;"]
    return [string map $map [join $args ""]]
}

proc xpath_str { str } {
    
    foreach "strList type pos" [list "" "" 0] break
    while 1 {
	switch $type {
	    "" {
		set ret [regexp -start $pos -indices {['"]} $str idxs]
		if { !$ret } {
		    lappend strList "\"[string range $str $pos end]\""
		    break
		}
		set idx [lindex $idxs 0]
		switch -- [string index $str $idx] {
		    ' { set type apostrophe }
		    \" { set type quote }
		}
	    }
	    apostrophe {
		set ret [regexp -start $pos -indices {["]} $str idxs]
		if { !$ret } {
		    lappend strList "\"[string range $str $pos end]\""
		    break
		}
		set idx [lindex $idxs 0]
		lappend strList "\"[string range $str $pos [expr {$idx-1}]]\""
		set type quote
		set pos $idx
	    }
	    quote {
		set ret [regexp -start $pos -indices {[']} $str idxs]
		if { !$ret } {
		    lappend strList "'[string range $str $pos end]'"
		    break
		}
		set idx [lindex $idxs 0]
		lappend strList "'[string range $str $pos [expr {$idx-1}]]'"
		set type apostrophe
		set pos $idx
	    }
	}
    }
    if { [llength $strList] > 1 } {
	return "concat([join $strList ,])"
    } else {
	return [lindex $strList 0]
    }
}

proc format_xpath { string args } {
    set cmd [list format $string]
    foreach i $args {
	lappend cmd [xpath_str $i]
    }
    return [eval $cmd]
}

namespace eval ::dom::domNode {}

# args can be one or more tags
proc ::dom::domNode::appendChildTag { node args } {
    if { [::llength $args] == 0 } {
	error "error in appendChildTag. At list one tag"
    }
    ::set doc [$node ownerDocument]
    foreach tag $args {
	if { [string match "text() *" $tag] } {
	    ::set newnode [$doc createTextNode [lindex $tag 1]]
	    $node appendChild $newnode
	    ::set node $newnode
	} elseif { [string match "attributes() *" $tag] } {
	    foreach "n v" [lrange $tag 1 end] {
		$node setAttribute $n $v
	    }
	} else {
	    ::set newnode [$doc createElement $tag]
	    $node appendChild $newnode
	    ::set node $newnode
	}
    }
    return $newnode
}

# args can be one or more tags
proc ::dom::domNode::appendChildTagNS { node url args } {
    if { [::llength $args] == 0 } {
	error "error in appendChildTagNS. At list one tag"
    }
    ::set doc [$node ownerDocument]
    foreach tag $args {
	if { [string match "text() *" $tag] } {
	    ::set newnode [$doc createTextNode [lindex $tag 1]]
	    $node appendChild $newnode
	    ::set node $newnode
	} elseif { [string match "attributes() *" $tag] } {
	    foreach "n v" [lrange $tag 1 end] {
		$node setAttribute $n $v
	    }
	} else {
	    ::set newnode [$doc createElementNS $url $tag]
	    $node appendChild $newnode
	    ::set node $newnode
	}
    }
    return $newnode
}

proc ::dom::domNode::appendChildText { node text } {
    ::set doc [$node ownerDocument]
    foreach child [$node selectNodes text()] { $child delete }
    ::set newnode [$doc createTextNode $text]
    $node appendChild $newnode
    return $newnode
}

proc ::dom::domNode::insertAfter { node newChild refChild } {

    ::set nextSibling [$refChild nextSibling]
    if { $nextSibling ne "" } {
	return [$node insertBefore $newChild $nextSibling]
    } else {
	return [$node appendChild $newChild]
    }
}

proc ::dom::domNode::insertBeforeTag { node tag refChild } {
    ::set doc [$node ownerDocument]
    ::set newnode [$doc createElement $tag]
    $node insertBefore $newnode $refChild
    return $newnode
}

proc ::dom::domNode::insertAfterTag { node tag refChild } {
    ::set doc [$node ownerDocument]
    ::set newnode [$doc createElement $tag]
    $node insertAfter $newnode $refChild
    return $newnode
}


################################################################################
#    compatibility functions
#
################################################################################

if { [package vcompare 8.5 [package present Tcl]] > 0 } {
    namespace eval ::cu::compatibility8.4 {}
    proc ::cu::compatibility8.4::go {} {
	namespace eval ::cu::compatibility8.4 {}
	rename ::lsearch ::cu::compatibility8.4::lsearch
	
	proc ::lsearch { args } {
	    set index -1
	    for { set i 0 } { $i < [llength $args] } { incr i } {
		switch -- [lindex $args $i] {
		    -index {
		        set index [lindex $args [expr {$i+1}]]
		        set args [lreplace $args $i [expr {$i+1}]]
		        break
		    }
		    -- { break }
		}
	    }
	    if { $index >= 0 } {
		set nlist ""
		foreach i [lindex $args end-1] {
		    lappend nlist [lindex $i $index]
		}
		lset args end-1 $nlist
		return [eval ::cu::compatibility8.4::lsearch $args]
	    } else {
		return [eval ::cu::compatibility8.4::lsearch $args]
	    }
	}
    }
    ::cu::compatibility8.4::go
}

if { [info command lrepeat] eq "" } {
    proc lrepeat {count value args} {
	set values [linsert $args 0 $value]
	set result {}
	for {set i 0} {$i < $count} {incr i} {
	    eval [list lappend result] $values
	}
	return $result
    }
}

################################################################################
#    store preferences
################################################################################

proc cu::store_program_preferences { args } {

    set optional {
	{ -valueName name "" }
    }
    set compulsory "program_name data"

    parse_args $optional $compulsory $args

    if { $valueName eq "" } {
	set valueNameF IniData
    } else {
	set valueNameF IniData_$valueName
    }

    if { $::tcl_platform(platform) == "windows" } {
	set key "HKEY_CURRENT_USER\\Software\\Compass\\$program_name"
	package require registry
	registry set $key $valueNameF $data
    } else {
	package require tdom
	set file [file normalize ~/.${program_name}_prefs]
	set err [catch { tDOM::xmlReadFile $file } xml]
	if { $err } { set xml "<preferences/>" }
	set doc [dom parse $xml]
	set root [$doc documentElement]
	set domNode [$root selectNodes "pref\[@n=[xpath_str $valueNameF]\]"]
	if { $domNode ne "" } { $domNode delete }
	set p [$root appendChildTag pref]
	$p setAttribute n $valueNameF
	$p appendChildText $data

	set fout [open $file w]
	fconfigure $fout -encoding utf-8
	puts $fout [$doc asXML]
	close $fout
    }
}
proc cu::get_program_preferences { args } {

    set optional {
	{ -valueName name "" }
    }
    set compulsory "program_name"

    parse_args $optional $compulsory $args

    if { $valueName eq "" } {
	set valueNameF IniData
    } else {
	set valueNameF IniData_$valueName
    }

    set data ""
    if { $::tcl_platform(platform) == "windows" } {
	set key "HKEY_CURRENT_USER\\Software\\Compass\\$program_name"
	package require registry
	catch { registry get $key $valueNameF } data
    } else {
	package require tdom
	set file [file normalize ~/.${program_name}_prefs]
	set err [catch { tDOM::xmlReadFile $file } xml]
	if { !$err } {
	    set doc [dom parse $xml]
	    set root [$doc documentElement]
	    set domNode [$root selectNodes "pref\[@n=[xpath_str $valueNameF]\]"]
	    if { $domNode ne "" } {
		set data [$domNode text]
	    }
	}
    }
    return $data
}

################################################################################
#    zip and unzip files
################################################################################

proc cu::zipfile { args } {

    set recursive 0
    set compression 9
    set isdata 0

    for { set i 0 } { $i < [llength $args] } { incr i } {
	set todel 0
	switch -regexp -- [lindex $args $i] {
	    {^-r$} { set recursive 1 ; set todel 1 }
	    {^-[0-9]$} { set compression [string trimleft [lindex $args $i] -] ; set todel 1 }
	    {^-data$} { set isdata 1 ; set todel 1 }
	    {^-} { error "unkown option '[lindex $args $i]'" }
	}
	if { !$todel } { break }
	set args [lreplace $args $i [expr {$i+$todel-1}]]
	incr i [expr {-1*$todel}]
    }
    if { $isdata && $recursive } {
	error "error: options -data and -r cannot be used together in cu::zipfile"
    }
    set zipfile [lindex $args 0]
    set files [lrange $args 1 end]

    if { $recursive } {
	for { set i 0 } { $i < [llength $files] } { incr i } {
	    if { [file isdirectory [lindex $files $i]] } {
		eval lappend files [glob -nocomplain -dir [lindex $files $i] *]
	    }
	}
    }
    
    set h [zip open $zipfile w]
    if { !$isdata } {
	foreach i $files {
	    if { [file isdirectory $i] } { continue }
	    zip set $h $i -level $compression -time [file mtime $i]
	    set fin [open $i r]
	    fconfigure $fin -translation binary
	    zip write $h [read $fin]
	    close $fin
	}
    } else {
	set now [clock seconds]
	foreach i $files {
	    foreach "name data" $i break
	    zip set $h $name -level $compression -time $now
	    zip write $h $data
	}
    }
    zip close $h
}

proc cu::unzipfile { args } {

    set overwrite ask
    set isdata 0

    for { set i 0 } { $i < [llength $args] } { incr i } {
	set todel 0
	switch -regexp -- [lindex $args $i] {
	    {^-o$} { set overwrite always ; set todel 1 }
	    {^-data$} { set isdata 1 ; set todel 1 }
	    {^-} { error "unkown option '[lindex $args $i]'" }
	}
	if { !$todel } { break }
	set args [lreplace $args $i [expr {$i+$todel-1}]]
	incr i [expr {-1*$todel}]
    }
    set zipfile [lindex $args 0]

    set retval ""
    set h [zip open $zipfile r]
    if { !$isdata } {
	foreach file [zip files $h] {
	    foreach "mtime csize size att comment" [zip info $h $file] break
	    if { ![file exists [file dirname $file]] } { file mkdir [file dirname $file] }
	    if { [file exists $file] && $overwrite != "always" } {
		if { $overwrite == "ask" } {
		    package require Tk
		    set retval [tk_dialog .d "Confirm" "Do you want to overwrite file '$file'?" \
		                    question 0 Yes "Yes to All" No "No to all" Cancel]
		    switch -- $retval {
		        4 - -1 {
		            zip close $h
		            return
		        }
		        0 {
		            #nothing
		        }
		        1 { set overwrite always }
		        2 { continue }
		        3 { set overwrite never ; continue }
		    }
		} elseif { $overwrite == "never" } { continue }
	    }
	    zip set $h $file
	    set fout [open $file w]
	    fconfigure $fout -translation binary
	    puts -nonewline $fout [zip read $h]
	    close $fout
	    file mtime $file $mtime
	    lappend retval [list $file $mtime]
	}
    } else {
	foreach file [zip files $h] {
	    foreach "mtime csize size att comment" [zip info $h $file] break
	    zip set $h $file
	    lappend retval [list $file $mtime [zip read $h]]
	}
    }
    zip close $h
    return $retval
}








